import os
import time
import json
import hmac
import base64
import random
import hashlib
import requests
from typing import Tuple


class Alibaba:
    def __init__(
        self,
        ACCESS_KEY_ID=os.environ.get("ALIBABA_CLOUD_ACCESS_KEY_ID", None),
        ACCESS_KEY_SECRET=os.environ.get("ALIBABA_CLOUD_ACCESS_KEY_SECRET", None),
        SECURITY_TOKEN=os.environ.get("ALIBABA_CLOUD_SECURITY_TOKEN", None),
        ACCOUNT_ID=os.environ.get("FC_ACCOUNT_ID", None),
        region=os.environ.get("FC_REGION", "cn-beijing"),
        logger=None,
    ):
        """Initialize the Alibaba class with the provided parameters.

        :param ACCESS_KEY_ID: Alibaba Cloud Access Key ID
        :param ACCESS_KEY_SECRET: Alibaba Cloud Access Key Secret
        :param SECURITY_TOKEN: Alibaba Cloud Security Token
        :param ACCOUNT_ID: Alibaba Cloud Account ID
        :param config: Configuration for the Alibaba class
        :param logger: Logger for the Alibaba class
        """
        self.logger = logger
        self.ALIBABA_CLOUD_ACCESS_KEY_ID = ACCESS_KEY_ID
        self.ALIBABA_CLOUD_SECURITY_TOKEN = SECURITY_TOKEN
        self.ALIBABA_CLOUD_ACCESS_KEY_SECRET = ACCESS_KEY_SECRET
        self.endpoint = f"{ACCOUNT_ID}.{region}.fc.aliyuncs.com"
    
    def sign_request(self, method, headers, resource):
        """Alibaba cloud api request sign method;
        docs: https://help.aliyun.com/zh/sdk/product-overview/roa-mechanism?spm=a2c4g.2618586.0.i9.

        :param method: request method
        :param headers: request headers
        :param resource: request uri
        :return: signed string, like: "acs keyid:sign".
        """
        string_to_sign = (
            f"{method}\n{headers.get('accept', '')}\n{headers.get('content-md5', '')}\n"
            f"{headers.get('content-type', '')}\n{headers.get('date', '')}\n"
            f"x-acs-accesskey-id:{self.ALIBABA_CLOUD_ACCESS_KEY_ID}\n"
            f"x-acs-security-token:{self.ALIBABA_CLOUD_SECURITY_TOKEN}\n"
            f"x-acs-signature-method:HMAC-SHA1\n"
            f"x-acs-signature-nonce:{headers.get('x-acs-signature-nonce', '')}\n"
            f"x-acs-signature-version:1.0\nx-acs-version:{headers.get('x-acs-version', '')}\n{resource}"
        )
        key_secret_encode = self.ALIBABA_CLOUD_ACCESS_KEY_SECRET.encode("utf-8")
        string_to_sign_encode = string_to_sign.encode("utf-8")
        signature = hmac.new(
            key_secret_encode, string_to_sign_encode, hashlib.sha1
        ).digest()
        signature = base64.b64encode(signature)
        return f"acs {self.ALIBABA_CLOUD_ACCESS_KEY_ID}:{signature.decode('utf-8')}"

    def get_headers(self):
        """Get default request headers, include "x-acs-signature" and "x-acs-version".

        :return: request headers.
        """
        return {
            "accept": "application/json",
            "date": time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime()),
            "host": self.endpoint,
            "x-acs-signature-nonce": str(random.randint(1, 10000)),
            "x-acs-signature-method": "HMAC-SHA1",
            "x-acs-signature-version": "1.0",
            "x-acs-version": "2023-03-30",
            "content-type": "application/json",
            "content-md5": "",
            "x-acs-security-token": self.ALIBABA_CLOUD_SECURITY_TOKEN,
            "x-acs-accesskey-id": self.ALIBABA_CLOUD_ACCESS_KEY_ID
        }

    def get_function(self, function_name):
        """Get the function detail;
        docs: https://help.aliyun.com/document_detail/2618610.html?spm=a2c4g.2618615.0.0.4b4613f35Qq27z.

        :param function_name: function name, default regex: dipperai-{model_platform}-{model_id}-{model_version}
        :return: the request response.
        """
        try:
            resource = f"/2023-03-30/functions/{function_name}"
            headers = self.get_headers()
            headers["authorization"] = self.sign_request(
                "GET", headers, resource
            )
            url = f"https://{self.endpoint}{resource}"
            resp = requests.get(url, headers=headers)
            response_data = resp.content.decode("utf-8")
            if resp.status_code != 200:
                self.logger.error(f"get function error: {response_data}")
                return {}
            return json.loads(response_data)
        except Exception as e:
            self.logger.error(e)
            return {}

    def get_trigger(self, function_name, trigger_name="dipperai_default_trigger"):
        """Get the function trigger detail;
        docs: https://help.aliyun.com/document_detail/2618615.html?spm=a2c4g.2618641.0.0.79653c17XYM3S8.

        :param function_name: function name, default regex: dipperai-{model_platform}-{model_id}-{model_version}
        :param trigger_name: trigger name, default: dipperai_default_trigger
        :return: the request response.
        """
        try:
            resource = (
                f"/2023-03-30/functions/{function_name}/triggers/{trigger_name}"
            )
            headers = self.get_headers()
            headers["authorization"] = self.sign_request(
                "GET", headers, resource
            )
            url = f"https://{self.endpoint}{resource}"
            resp = requests.get(url, headers=headers)
            response_data = resp.content.decode("utf-8")
            if resp.status_code != 200:
                self.logger.error(f"get trigger error: {response_data}")
                return {}
            return json.loads(response_data)
        except Exception as e:
            self.logger.error(e)
            return {}

    def create_function(self, function_name:str, function_config:dict):
        """Create alibaba cloud fc function, default is custom container function;
        docs: https://help.aliyun.com/document_detail/2618641.html?spm=a2c4g.2618639.0.0.393b7c57jVkKrX.

        :param function_name: function name, default regex: dipperai-{model_platform}-{model_id}-{model_version}
        :param function_config: function config
        :return: created response.
        """
        try:
            resource = "/2023-03-30/functions"
            headers = self.get_headers()
            headers["authorization"] = self.sign_request(
                "POST", headers, resource
            )
            url = f"https://{self.endpoint}{resource}"
            default_config = {
                "cpu": 0.05,
                "customContainerConfig": {},
                "description": "Serverless AI Project, {model_id}-{model_version}",
                "environmentVariables": {},
                "functionName": function_name,
                "gpuConfig": {},
                "role": "",
                "runtime": "custom-container",
                "timeout": 300,
            }
            merged_config = {**default_config, **(function_config or {})}
            resp = requests.post(
                url, headers=headers, json=merged_config
            )
            response_data = resp.content.decode("utf-8")
            if resp.status_code != 200:
                self.logger.error(f"create function error {response_data}")
                return {}
            return json.loads(response_data)
        except Exception as e:
            self.logger.error(e)
            return {}

    def create_trigger(self, function_name, trigger_name="dipperai_default_trigger"):
        """Create alibaba cloud fc function trigger;
        docs: https://help.aliyun.com/document_detail/2618639.html?spm=a2c4g.2508973.0.0.3d7c7c57JwPHEM.

        more: the trigger config detail: https://github.com/devsapp/fc/blob/main/src/lib/interface/fc/trigger.ts
        :param function_name: function name, default regex: serverlessas-{model_id}-{model_version}
        :param trigger_name: trigger name, default: dipperai_default_trigger
        :return: created response.
        """
        try:
            resource = f"/2023-03-30/functions/{function_name}/triggers"
            headers = self.get_headers()
            headers["authorization"] = self.sign_request(
                "POST", headers, resource
            )
            url = f"https://{self.endpoint}{resource}"
            resp = requests.post(
                url,
                headers=headers,
                json={
                    "description": "Serverless AI Project Default HTTP Trigger",
                    "qualifier": "LATEST",
                    "triggerConfig": json.dumps({
                        "authType": "anonymous",
                        "methods": ["GET", "POST"],
                    }),
                    "triggerName": trigger_name,
                    "triggerType": "http",
                },
            )
            response_data = resp.content.decode("utf-8")
            if resp.status_code != 200:
                self.logger.error(f"create trigger {response_data}")
                return {}
            return json.loads(response_data)
        except Exception as e:
            self.logger.error(e)
            return {}

    def check(self, name: str) -> Tuple[dict, bool]:
        """
        Check function is exist.

        :param name: function name
        :return: check result.
        """
        try:
            function_info = self.get_function(function_name=name)
            if function_info:
                trigger_info = self.get_trigger(function_name=name)
                function_info["triggers"] = trigger_info
                return {
                    "config": function_info,
                    "url": trigger_info["httpTrigger"]["urlInternet"],
                }, True
        except Exception as e:
            self.logger.error(e)
        return {}, False

    def update(self, name:str, config:dict) ->Tuple[dict, bool]:
        """
        Update the function to the specify config.

        :param name: function name
        :param config: function config
        :return: update result.
        """
        pass

    def create(self, name: str, config: dict = None) -> dict:
        """
        Create the function with the specify config.

        :param name: function name
        :param config: function config
        :return: create result.
        """
        try:
            fc_config = self.create_function(function_name=name, function_config=config)
            if fc_config:
                trigger = self.create_trigger(function_name=name)
                self.logger.info(f"deploy function {name} to alibaba cloud function compute")
                return {
                    "url":trigger["httpTrigger"]["urlInternet"],
                    "config": fc_config
                }
        except Exception as e:
            self.logger.error(e)
            return {"error" : e}
        return {}



import importlib
import requests
from ..vendor import Alibaba
from ..utils.logger import setup_logger
from ..utils.cache import OperateCache, cache

model_default_error = "Model service exception."
model_check_error = "Model service check failed."
model_update_error = "Model service update failed."
model_create_error = "Model service create failed."


class MaaS:

    def __init__(self, model_url=None, model_id=None, model_version="master", cloud=None, service_config=None,
                 service_url=None, debug=False):
        """
        init default vars
        :param model_url: model url, like: https://modelscope.cn/models/iic/cv_convnextTiny_ocr-recognition-general_damo/summary
        :param model_id: model name / model id
        :param model_version: model version
        :param cloud: cloud/vendor attr, like: Alibaba(), AWS(), AIGCaaS()
        :param service_config: function/container config
        :param service_url: service url
        :param debug: debug mode
        """
        self.default_resource_name = None
        self.service_url = service_url
        self.model_id = model_id
        self.model_url = model_url
        self.model_version = model_version
        self.maas_name = self.__class__.__name__
        self.logger = setup_logger(debug=debug)
        self.vendor = cloud or Alibaba(logger=self.logger)
        self.model_meta = self.get_model_meta()
        self.resource_name = self.get_resource_name()
        self.service_config = self.get_service_config(service_config)
        # 部署模型
        if not self.__deploy():
            raise Exception("Failed to deploy model to cloud.")
        else:
            self.logger.info("Model deployed successfully.")

    
    def get_config_func_name(self) -> str:
        """
        get resource name
        :return: str
        """
        return str("%s_%s_default_config" % (self.maas_name, self.vendor.__class__.__name__)).lower()

    def get_service_config(self, user_config: dict) -> dict:
        """
        get service config
        :return: dict
        """
        # 补充配置
        config_func_name = self.get_config_func_name()
        lib_module = importlib.import_module('DipperAI_inner.resources.config')
        return getattr(lib_module, config_func_name)(user_config)

    def get_model_meta(self) -> dict:
        """
        the MaaS module needs to rewrite this method, primarily for handling the model_url.
        It involves parsing the model_url to extract the model_id and model_version.
        :return: dict
        """
        return {}

    def check_config(self, new:dict, old:dict) -> bool:
        """
        compare the current configuration with the user specified configuration
        :param sub_dict: user newly specified configuration
        :param super_dict: current configuration, or default configuration
        :return: comparison results, True / False
                    True: indicates consistency
                    False: indicates inconsistency (requires update operation)
        """
        for key, value in new.items():
            if key not in old:
                print(key)
                return False
            if isinstance(value, dict):
                if not isinstance(old[key], dict) or not self.check_config(value, old[key]):
                    print(key)
                    return False
            elif old[key] != value: 
                    print(key)
                    return False
        return True
    
    def get_resource_name(self) -> str:
        """
        get resource name
        :return: resource name
        """
        if self.default_resource_name:
            return self.default_resource_name
        resource_name = f'dipperai-{self.maas_name}-{self.model_id}-{self.model_version}'
        # special attention should be paid to whether the name will exceed the maximum value
        # for example, a length of 64 is the upper limit
        return resource_name.replace("/", "-")

    def __deploy(self) -> str:
        """
        get service url; the main func of maas
        :return: service url
        """
        if self.service_url: 
            return True

        # get cache
        #   1. get cache success:
        #     1.1 config mismatching: update
        #       1.1.1 update success: return
        #       1.1.2 update failed: raise exception
        #       1.1.3 update unknown: raise exception
        #     1.2 config matching: return
        #   2. get cache failed: continue
        with OperateCache(cache) as cache_data:
            cache_model_info = cache_data.get_cache(self.resource_name)
            if cache_model_info and cache_model_info.get("url"):
                if not self.check_config(self.service_config, cache_model_info.get("config", {})):
                    update_result, is_ok = self.vendor.update(self.resource_name, self.service_config)
                    if not is_ok: 
                        raise BaseException(update_result.get("error", model_update_error))
                    # set_cache(self.resource_name, update_result)
                self.service_url = cache_model_info["url"]
                return True
            
            # check resources
            #   1. check resources success: return
            #     1.1 config mismatching: update
            #       1.1.1 update success:
            #       1.1.2 update failed:
            #       1.1.3 update unknown:
            #     1.2 config matching: return
            #   2. check resources failed: raise exception
            #   3. check resources unknown: continue
            check_result, existed = self.vendor.check(self.resource_name)
            if existed and check_result.get("url"):
                if not self.check_config(self.service_config, check_result.get("config")):
                    update_result, is_ok = self.vendor.update(self.resource_name, self.service_config)
                    if not is_ok or "config" not in update_result: 
                        raise BaseException(update_result.get("error", model_update_error))
                    self.service_config = update_result["config"]
                self.service_url = check_result["url"]
                cache_data.set_cache(self.resource_name, {
                    "url" : self.service_url,
                    "config":self.service_config
                })
                return True
            elif existed: 
                raise BaseException(model_check_error)
            
            # create resources
            #    1. create resources success: return
            #    2. create resources failed: raise exception
            #    3. create resources unknown: raise exception
            create_result = self.vendor.create(self.resource_name, self.service_config)
            if create_result.get("url"):
                cache_data.set_cache(self.resource_name, create_result)
                self.service_url = create_result["url"]
                self.service_config = create_result["config"]
                return True
            elif create_result.get("error"): 
                raise BaseException(create_result.get("error", model_create_error))
            return False
    
    def invoke(self, input: any, headers: dict = None) -> dict:
        """
        invoke MaaS
        :param input: input data
        :return: MaaS output
        """
        return requests.post(self.service_url, json=input, headers=headers).json()



"""The current file is used to configure default resource properties/configurations.

func name regex: {maas_class_name}_{vendor_class_name}_default_config (all letters in lowercase)
  like: modelscope_alibaba_default_config
:param: name: default resource name
:param: config: user specified configuration
:return: specific configuration information that will be used to create resources.
"""
import os

def modelscope_alibaba_default_config(config, **kwargs) -> dict:
    """Get modelscope alibaba function config.

    :param name: function name
    :param config: user config
    :return: config.
    """
    default_config = {
        "cpu": 0.05,
        "customContainerConfig": {},
        "description": "Serverless AI Project, {model_id}-{model_version}",
        "environmentVariables": {},
        "gpuConfig": {
        },
        "role": "",
        "runtime": "custom-container",
        "timeout": 300,
    }
    return

def alibaba_huggingface_default_config(name:str, model_id:str, access_token:str, model_task:str, library:str) -> dict:
        """
        获取默认函数配置
        """
        region = os.environ.get("FC_REGION", "cn-beijing")
        transformer = f"registry.{region}.aliyuncs.com/aliyun-fc/huggingface:transformers-v1"
        diffusers = f"registry.{region}.aliyuncs.com/aliyun-fc/huggingface:diffusers-v1"
        image = transformer if library == "transformers" else diffusers
        return {
            "functionName": name,
            "description": "huggingface model",
            "handler": "not-used",
            "timeout": 1800,
            "memorySize": 32768,
            "cpu":2,
            "diskSize": 10240,
            "gpuConfig":{
                "gpuMemorySize":16384,
                "gpuType":"fc.gpu.tesla.1"
            },
            "instanceConcurrency":1,
            "runtime": "custom-container",
            "internetAccess":True,
            "customContainerConfig":{
                "image": image,
                "port": 8000
            },
            "environmentVariables": {
                "MODEL_ID": model_id,
                "HUGGING_FACE_HUB_TOKEN": access_token,
                "MODEL_TASK": model_task,
                "HF_HOME": "/mnt/auto/hf",
                "PYTHONPATH": "/docker:/mnt/auto/python",
                "HF_ENDPOINT": "https://hf-mirror.com"
            }
        }

def huggingface_alibaba_default_config(config, **kwargs) -> dict:
    """
    返回一个合并了默认配置和用户配置的字典。

    参数:
    - name: 模型名称，当前未使用但可能用于未来扩展。
    - config: 用户提供的配置字典，如果未提供则默认为空字典。
    - **kwargs: 可变的关键字参数，可用于未来的扩展。

    返回值:
    - 一个字典，包含默认配置和用户配置的合并结果。
    """
    return {**alibaba_huggingface_default_config(**kwargs), **(config or {})}
    



import json
import requests
from maas.core import MaaS
from version import __version__


class Modelscope(MaaS):
    def __init__(
        self,
        model_url=None,
        model_id=None,
        model_version="master",
        cloud=None,
        config=None,
        url=None,
        debug=False,
    ):
        """Initialize the Modelscope instance.

        :param model_url: The URL of the model.
        :param model_id: The ID of the model.
        :param model_version: The version of the model, default is "master".
        :param cloud: The cloud provider.
        :param config: The configuration for the model.
        :param url: The URL for the model.
        :param debug: A flag indicating whether to run in debug mode.
        """
        super().__init__(
            model_url, model_id, model_version, cloud, config, url, debug
        )
        self.headers = {
            "Content-Type": "application/json",
            "User-Agent": "dipperai@%s" % __version__,
        }
        self.token = None

    def login(self):
        """
        login to modelscope, get cookie and add it to self.headers;
        source code: https://modelscope.cn/docs/api_docs/API文档%2Fbuild%2Fjson%2F_modules%2Fmodelscope%2Fhub%2Fapi%2F%23HubApi.login
        :return:
        """
        login_url = "https://modelscope.cn/api/v1/login"
        payload = json.dumps({"AccessToken": self.token})
        response_attr = requests.request("POST", login_url, headers=self.headers, data=payload)
        self.headers["Cookie"] = response_attr.cookies

    def get_task(self):
        """
        get modelscope model task type, automatically generated in file configuration.json;
        source code: https://modelscope.cn/docs/api_docs/API文档%2Fbuild%2Fjson%2F_modules%2Fmodelscope%2Fhub%2Ffile_download%2F%23model_file_download
        :return: Task type (https://modelscope.cn/docs/模型的推理Pipeline#当前支持的task列表)
        """
        file_url = 'https://modelscope.cn/api/v1/models/%s/repo?Revision=%s&FilePath=configuration.json' % (
            self.model_id, self.model_version)
        if self.token: self.login()
        response_json = json.loads(requests.request("POST", file_url, headers=self.headers).content.decode("utf-8"))
        return response_json["task"]
